//---------------------------------------------------------------------------
// GridBasedSpatialDisperse
//---------------------------------------------------------------------------
#if !defined(GridBasedSpatialDisperse_H)
  #define GridBasedSpatialDisperse_H

#include "DisperseBase.h"

class xercesc::DOMDocument;
class clPlot;
class clModelMath;
/**
  Anisotropic spatial dispersal - Version 1.0
  <p>
  This is a base class for behaviors that create and disperse seeds according
  to the spatially-explicit model.  For each grid cell, the number of seeds is
  calculated based on the size and distance of parents of that species.
  <p>
  This class and its descendents do not take gap status into account.
  Behaviors of this class cannot be applied to stumps.
  <p>
  The "seeds" are actually tallies by species in a grid.  This behavior does
  not produce any seedlings.  A separate recruitment behavior must "germinate"
  the seeds into seedlings according to its own rules.
  <p>
  Copyright 2003 Charles D. Canham.
  @author Lora E. Murphy

  <br>Edit history:
  <br>-----------------
  <br>April 28, 2004 - Submitted as beta (LEM)
*/
class clGridBasedSpatialDisperse : virtual public clDisperseBase {
//note: need the virtual keyword to avoid base class ambiguity.

 public:

/**
  Constructor.

  @param p_oSimManager Sim Manager object.
*/
 clGridBasedSpatialDisperse(clSimManager *p_oSimManager);

/**
  Destructor.
*/
 ~clGridBasedSpatialDisperse();

 protected:

 clPlot *mp_oPlot; /**<Plot object.*/
 float *mp_fDispersalX0;/**<Dispersal (weibull) or X0 (lognormal) parameter for
       each species.  This value comes from the parameter file.*/
 float *mp_fThetaXb;    /**<Theta (weibull) or Xb (lognormal) parameter for each
       species.  This value comes from the parameter file.*/
 float *mp_fStr;  /**<Seedling Total Recruits for each species.  This value
       comes from the parameter file.*/
 float *mp_fBeta; /**<Beta parameter for each species.  This value
       comes from the parameter file.*/
 float *mp_fFecundity;  /**<Fecundity for 30 cm dbh tree for each species.
       This value is calculated in CalculateFecundity().*/
 float m_fMaxSearchRadius; /**<Max distance in m to search for dispersing
       parents from the center of a grid cell.  This must be less than half the
       shortest plot distance.*/
 float m_fQueryHeight; /**<Shortest possible reproducing neighbor height, for
       putting in distance/height queries*/
 float m_fSeedCellArea; /**<Area of seed grid cell in square meters*/
 float m_fNumYearsPerTimestep; /**<Number of years per timestep*/
 short int *mp_iIndexes;  /**<This will speed access to the other arrays by
           storing each species' array index so the other arrays only have to
           be as big as the number of unique species for this behavior.*/
 short int m_iNumFunctions; /**<How many functions there are - 2*/
 short int *mp_iWhatFunction; /**<Which function is used for each species*/

  bool **mp_bIsUsed; /**<Whether this behavior is used by a species/type
       combo.  First array index is species, second is type.  This helps
       quickly sort through the results of neighbor searches.*/

/**
* Performs dispersal.  For each grid cell, the center point of the cell is
* calculated.  For each species to which this behavior applies, a search is
* done to find all trees within the maximum search radius.  For each tree,
* the number of seeds to add is gotten by multiplying the fecundity by the
* return value of the virtual function GetFunctionValue().  The total number
* of seeds for a species is RandomRound()ed to arrive at an integer.
*/
void AddSeeds();


/**
* Child classes can override this to do extra setup.
* @param p_oDoc Pointer to parsed parameter file.
*/
  virtual void DoChildSetup(xercesc::DOMDocument *p_oDoc) {;};

  /**
  Reads in values from the parameter file.

  @param p_oDoc DOM tree of parsed input file.
*/
 void DoShellSetup(xercesc::DOMDocument *p_oDoc);

  /**
  * Gets normalized function value which, when multiplied by fecundity, gets
  * the number of seeds for a tree.  Must be overridden by child classes.
  * @param fCellX X coordinate of point at center of grid cell into which
  * seeds are being dispersed.
  * @param fCellY Y coordinate of point at center of grid cell into which
  * seeds are being dispersed.
  * @param fTreeX X coordinate of neighbor parent tree.
  * @param fTreeY Y coordinate of neighbor parent tree.
  * @param iSpecies Species of neighbor parent tree.
  * @return Function value.
  */
  virtual float GetFunctionValue(float &fCellX, float &fCellY, float &fTreeX,
  float &fTreeY, int &iSpecies) = 0;

  /**
  * Declares the arrays.
  */
  void DeclareArrays();

  /**
  * Extracts needed parameter file data.  This function takes into account
  * whether this is isotropic or anisotropic.
  * @param p_oDoc Parsed parameter file document.
  * @throws modelErr if the function values are not valid, or if the max search
  * radius is negative or greater than half the shortest plot length.
  */
  void GetParameterFileData(xercesc::DOMDocument *p_oDoc);

  /**
  * Calculates fecundity.  Fecundity
  * is calculated as <br>
  * <center>fec = STR/30<sup>beta</sup></center>
  */
  void CalculateFecundity();

  /**
  * Declares and populates the mp_bIsUsed array.
  * @throws modelErr if the object is being applied to a type other than
  * sapling and adult.
  */
  void PopulateUsedTable();

  /**
  * Calculates the shortest possible reproducing neighbor, by finding the
  * shortest height at a species' min. reproductive DBH.
  */
  void CalculateQueryHeight();
};
//---------------------------------------------------------------------------
#endif // AnisotropicDisperse_H
