//---------------------------------------------------------------------------

#ifndef SelfThinningH
#define SelfThinningH
//---------------------------------------------------------------------------
#include "MortalityBase.h"
/**
 * Self-Thinning Mortality - Version 1.0
 * 
 * This evaluates self-thinning mortality, either adult or juvenile.  The one
 * that is done depends on which behavior name string is put in the parameter
 * file - either "adultselfthin" or "juvselfthin".  This controls which set of
 * parameters is read in from the parameter file.  The self-thinning is
 * calculated the same way either way.  There may be more than one object of 
 * this class running around.
 * 
 * If this is for adult self-thinning, a maximum dbh applies.  This is not true
 * of juvenile self-thinning, so be careful when writing the parameter file.
 * 
 * Despite the names, there is no restriction that says that adult self-thinning
 * can be applied only to adults, or juvenile self-thinning can be applied only
 * to juveniles.
 * 
 * This class's namestring is "selfthinmortshell".
 * 
 * Self-thinning mortality departs from our traditional modeling approach in 
 * that parameters are 'tuned' to site data for specific conditions such as 
 * highly dense, young spruce-aspen stands. Self-thinning is a pseudo-density 
 * dependent function that can overcomes the shortcoming of Sortie in killing 
 * small trees in highly dense uniform aged stands. This approach is applicable
 * only in these specific conditions. It is a heuristic model that should be 
 * replaced with a generalized multi-species self-thinning model when the data 
 * become available.
 * 
 * Copyright 2003 Charles D. Canham.
 * @author Lora E. Murphy
 * 
 * <br>Edit history:
 * <br>-----------------
 * <br>April 28, 2004 - Submitted as beta (LEM)
 * <br>February 8, 2008 - Changed mortality flags from simple booleans to reason
 * codes (LEM)
 */

class clSelfThinMort : virtual public clMortalityBase {
  //note: need the virtual keyword to avoid base class ambiguity.

public:

  /**
   * Constructor.  Sets the namestring and NULLs pointers.
   * @param p_oSimManager Sim Manager object.
   */
  clSelfThinMort(clSimManager *p_oSimManager);

  /**
   * Destructor.  Deletes arrays.
   */
  ~clSelfThinMort();

  /**
   * Reads in values from the parameter file.
   * @param p_oDoc DOM tree of parsed input file.
   */
  void DoShellSetup(xercesc::DOMDocument *p_oDoc);

  /**
   * Calculates mortality according to the self-thinning function.
   *
   * @param p_oTree Tree being evaluated
   * @param fDiam DBH of tree being evaluated - for seedlings will be 0
   * @param iSpecies Species of the tree being evaluated
   * @return natural if the tree is to die, notdead if it lives.
   */
  whyDead DoMort(clTree *p_oTree, const float &fDiam, const short int &iSpecies);

  /**
   * Captures the behavior name passed from the parameter file.  This is useful
   * since this class can produce a few different kinds of behaviors.
   * @param cNameString Behavior name from parameter file.
   */
  void SetNameData(char *cNameString);

protected:
  float *mp_fSelfThinSlope; /**<Slope of self-thinning function*/
  float *mp_fSelfThinIntercept; /**<Intercept of self-thinning function*/
  float *mp_fSelfThinMaxDbh; /**<Max DBH for self-thinning.  Adult only.*/
  float m_fNumberYearsPerTimestep; /**<Number years/timestep-from sim manager*/

  bool m_bIsAdult; /**<If true - adult self-thinning - if false, juvenile.*/

};
//---------------------------------------------------------------------------
#endif
